from typing import Any, List, Dict, Mapping, Text, Optional, Tuple, TypeVar, Union, Awaitable
from datetime import date, datetime, timedelta
import logging
import redis
from ewoxdbredis.repositories.redis_base_repository import RedisBaseRepository

T = TypeVar("T", redis.asyncio.RedisCluster, redis.asyncio.Redis)


class RedisHashRepository(RedisBaseRepository):
    def __init__(self, table_name:str="") -> None:
        super().__init__(table_name)


    async def set(self, connection:T, key:str, value:str) -> bool:
        """ Set a key/value without expiration. """
        if (self._table_name):
            await connection.hset(self._table_name, key, value)
            return True
        
        return False


    async def set_expire(self, connection:T, key:str, value:str, expire_at:datetime) -> bool:
        """ Add a key/value that expires in x-seconds. """
        if (self._table_name):
            await connection.hset(self._table_name, key, value)
            await connection.expireat(key, expire_at)
            return True
        
        return False


    async def get(self, connection:T, key:str) -> str:
        """ Get a value by its key. If the key does not exist, return an empty string. """
        if (self._table_name):
            value:str = await connection.hget(self._table_name, key)
            return value.decode() if (value is not None) else ""
        
        return ""


    async def set_item(self, connection:T, key:str, item:Dict) -> bool:
        """ Set a hash item in the Redis store. """
        env_key:str = self.get_enviroment_prefix() + key
        return await connection.hset(env_key, mapping=item)


    async def set_item_expire_at(self, connection:T, key:str, item:Dict, expire_at:datetime) -> bool:
        """ Set a hash item in the Redis store with an expiration time. """
        env_key:str = self.get_enviroment_prefix() + key
        await connection.hset(env_key, mapping=item)
        await connection.expireat(key, expire_at)
        return True


    async def get_item(self, connection:T, key:str, use_env:bool=True) -> Dict:
        """ Get a hash item by its key. If the key does not exist, return an empty dictionary. """
        env_key:str = self.get_enviroment_prefix() + key if (use_env) else key
        data:Dict = await connection.hgetall(env_key)
        model:Dict = dict()
        for key, value in data.items():
            model[key.decode()] = value.decode()

        return model


    async def get_keys_by_filter(self, connection:T, match:str) -> List[str]:
        """ Get all keys that match a specific pattern. """
        keys:List[str] = []
        env_match:str = self.get_enviroment_prefix() + match
        for key in await connection.scan_iter(match=env_match):
            keys.append(key.decode())

        return keys


    async def delete_item(self, connection:T, key:str) -> int:
        """ Delete a hash item by its key. Returns the number of deleted fields. """
        env_key:str = self.get_enviroment_prefix() + key
        return await connection.hdel(env_key)


    async def get_items(self, connection:T, keys:List[str]) -> List[Tuple[str, str]]:
        """ Get multiple items by their keys. Returns a list of tuples (key, value). """
        num_keys:int = len(keys)
        if (num_keys == 0):
            return []
        if (self._table_name):
            items:List[Tuple[str, str]] = []
            items_internal:List[str] = await connection.hmget(self._table_name, keys)
            if (items_internal is not None):
                for index, val in enumerate(items_internal):
                    if (val is not None):
                        if (index < num_keys):
                            items.append((keys[index], val.decode()))
            return items
        
        return []


    async def delete(self, connection:T, key:str) -> bool:
        """ Delete a key from the Redis hash table. Returns True if the key was deleted, False otherwise. """
        if (self._table_name):
            await connection.hdel(self._table_name, key)
            return True
        
        return False


    async def exist(self, connection:T, key:str) -> bool:
        """ Check if a key exists in the Redis hash table. Returns True if the key exists, False otherwise. """
        if (self._table_name):
            return await connection.hexists(self._table_name, key)
        
        return False


    async def get_num(self, connection:T) -> int:
        """ Get the number of items in the Redis hash table. """
        if (self._table_name):
            return await connection.hlen(self._table_name)

        return 0


    async def increment_by(self, connection:T, key:str, value:int) -> int:
        """ Increment the value of a key in the Redis hash table by a specified amount. """
        new_value:int = await connection.hincrby(self._table_name, key, value)
        return new_value


    async def get_table_keys(self, connection:T) -> List[str]:
        """ Get all keys in a hash table. Do not use this function on production. """
        keys_decoded:List[str] = []
        if (self._table_name):
            keys = await connection.hkeys(self._table_name)
            for key in keys:
                keys_decoded.append(key.decode())

        return keys_decoded


if __name__ == "__main__":
    from ewoxdbredis.connection.db_connection import DBConnection

    rep = RedisHashRepository()
    rep.set_table_name("cache")
    with DBConnection() as connection:
        expire_at:datetime = datetime.now() + timedelta(seconds=20)
        rep.set_expire(connection, "cacheKey1", "123", expire_at)

    rep = RedisHashRepository()
    rep.set_table_name("users")
    with DBConnection() as connection:
        rep.set_table_name("users")
        keys:List[str] = rep.get_keys(connection)
        user:Dict = dict()
        user["id"] = 1
        user["name"] = "Dude"
        user["video_likes"] = "user:1:video_likes"
        rep.set_item(connection, "user:1", user)

        video_likes:Dict = dict()
        video_likes["video1"] = 100
        video_likes["video2"] = 100
        video_likes["video3"] = 90
        rep.set_item(connection, "user:1:video_likes", video_likes)
        item = rep.get_item(connection, "user:1:video_likes")
        rep.set(connection, "123", 100)
        rep.set(connection, "124", 99)
        rep.set(connection, "125", 101)
        rep.set(connection, "126", 9)

        rep.get_keys_by_filter(connection, "user:*")

        value:str = rep.get(connection, "124")
        keys:List[str] = []
        keys.append("124")
        keys.append("126")
        items = rep.get_items(connection, keys)
        print(items)



from typing import Any, List, Dict, Mapping, Text, Optional, Tuple, TypeVar, Union, Awaitable
import logging
import redis
import os
from redis.client import Pipeline

T = TypeVar("T", redis.asyncio.RedisCluster, redis.asyncio.Redis)


class RedisBaseRepository():
    def __init__(self, table_name:str="") -> None:
        self._table_name:str = ""
        if (table_name != ""):
            self.set_table_name(table_name)


    def set_table_name(self, table_name:str) -> None:
        """ Set the table name with the environment prefix. """
        self._table_name = self.get_enviroment_prefix() + table_name


    def get_table_name(self) -> str:
        """ Get the table name with the environment prefix. """
        return self._table_name


    def get_enviroment_prefix(self) -> str:
        """ Get the environment prefix from the environment variable APP_ENV. """
        env_prefix:str = os.getenv("APP_ENV").lower() + "_"
        return env_prefix


    async def get_keys(self, connection:T) -> List[str]:
        """ Get all keys from the Redis connection. """
        keys = await connection.keys()
        keys_decoded:List[str] = []
        if (keys):
            for key in keys:
                keys_decoded.append(key.decode())

        return keys_decoded


    async def get_pipeline(self, connection:T) -> Pipeline:
        """ NOTE: Using the Pipeline requires extra responsibility. 
                  All keys must be prefixed with the environment. """
        return await connection.pipeline()



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
from abc import ABC, abstractmethod
import redis

T = TypeVar("T", redis.asyncio.RedisCluster, redis.asyncio.Redis)


class IRedisClient(ABC):
    @abstractmethod
    def get_connection(self) -> T:
        """ Get a Redis connection. """
        raise NotImplementedError("Implement inhereted method")
    
    @abstractmethod
    async def setup(self) -> None:
        """ Setup the Redis connection. """
        raise NotImplementedError("Implement inhereted method")   

    @abstractmethod
    async def dispose(self) -> None:
        """ Dispose the Redis connection. """
        raise NotImplementedError("Implement inhereted method")
    

    @abstractmethod
    async def is_up(self) -> bool:
        """ Check if the Redis connection is up. """
        raise NotImplementedError("Implement inhereted method")



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from datetime import date, datetime, timedelta
from ewoxcore.utils.json_util import JsonUtil
from ewoxcore.cache.cache import Cache
from ewoxcore.cache.icache_provider import ICacheProvider
from redis.asyncio import client
from ewoxdbredis.clients.iredis_client import IRedisClient
from ewoxdbredis.connection.db_connection import DBConnection
from ewoxdbredis.repositories.redis_expire_repository import RedisExpireRepository

T = TypeVar("T")


class DistributedCacheProvider(ICacheProvider):
    def __init__(self, client:IRedisClient):
        self._repository:RedisExpireRepository = RedisExpireRepository()
        self._repository.set_table_name("cache")
        self._cache = Cache()
        self._use_local:bool = False
        self._client:IRedisClient = client


    def init(self, use_local:bool) -> None:
        """ Initialize the cache provider."""
        self._use_local = use_local


    def change_use_local(self, use_local:bool) -> None:
        """ Change the use of local cache."""
        self._use_local = use_local


    def activate_expiration(self, timer_interval:int) -> None:
        """ Activate expiration for the cache."""
        if (self._use_local):
            self._cache.activate_expiration(timer_interval)


    def set_cache_size(self, size:int) -> None:
        """ Set the size of the cache."""
        if (self._use_local):
            self._cache.set_cache_size(size)


    async def get(self, cacheKey:str) -> Any:
        """ Get a value from the cache by its key."""
        value:Any = None
        if (self._use_local):
            value = self._cache.get(cacheKey)
            if (value is not None):
                return value

        return await self._repository.get(self._client.get_connection(), cacheKey)


    async def insert(self, cacheKey:str, cacheValue:Any) -> None:
        """ Insert a value into the cache with the given key."""
        if (self._use_local):
            self._cache.insert(cacheKey, cacheValue)

        await self._repository.set(self._client.get_connection(), cacheKey, cacheValue)


    async def remove(self, cacheKey:str) -> None:
        """ Remove a value from the cache by its key."""
        if (self._use_local):
            self._cache.remove(cacheKey)

        await self._repository.delete(self._client.get_connection(), cacheKey)


    def get_all_values(self) -> List[Any]:
        """ Get all values from the cache."""
        raise NotImplementedError("Not supported.")


    async def add(self, key:str, value:str, absolute_expiration:timedelta) -> None:
        """ Add a value to the cache with an absolute expiration."""
        if (self._use_local):
            self._cache.add(key, value, absolute_expiration)

        await self._repository.set_expire(self._client.get_connection(), key, value, absolute_expiration)


    async def add_expire_at(self, key:str, value:str, expire_at:datetime) -> None:
        """ Add a value to the cache with an expiration at a specific datetime."""
        if (self._use_local):
            self._cache.add_expire_at(key, value, expire_at)

        await self._repository.set_expire_at(self._client.get_connection(), key, value, expire_at)


    async def clear(self) -> None:
        """ Clear local cache L1 """
        self._cache.clear()


    async def add_ext(self, key:str, value:T, absolute_expiration:timedelta) -> None:
        """ Add a value to the cache with an absolute expiration, serialized as Base64 encode JSON."""
        value_internal:str = JsonUtil.serializeJson64(value)
        await self.add(key, value_internal, absolute_expiration)


    async def add_expire_at_ext(self, key:str, value:T, expire_at:datetime) -> None:
        """ Add a value to the cache with an expiration at a specific datetime, serialized as Base64 encode JSON."""
        value_internal:str = JsonUtil.serializeJson64(value)
        await self.add_expire_at(key, value_internal, expire_at)


    async def insert_ext(self, key:str, value:T) -> None:
        """ Insert a value into the cache with the given key, serialized as Base64 encode JSON."""
        value_internal:str = JsonUtil.serializeJson64(value)
        await self.insert(key, value_internal)


    async def get_ext(self, class_type:T, key:str) -> T:
        """ Get a value from the cache by its key, deserialized from Base64 encode JSON."""
        value_enc:str = await self.get(key)
        return JsonUtil.deserialize_json64(class_type, value_enc)



if __name__ == "__main__":
    prov:DistributedCacheProvider = DistributedCacheProvider()
    prov.add("cacheKey2", "Dude has a car!", timedelta(seconds=20))
    val:str = prov.get("cacheKey2")
    print(val)


from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type, cast
#import redis
from redis.asyncio import Redis, RedisCluster, ResponseError
import asyncio
from redis.exceptions import ConnectionError, TimeoutError
from tenacity import retry, wait_exponential, wait_fixed, stop_after_attempt, RetryError
from ewoxdbredis.clients.iredis_client import IRedisClient
from ewoxdbredis.settings.consumer_settings import ConsumerSettings
from ewoxcore.message.imessage_consumer import IMessageConsumer
from ewoxcore.message.message_args import MessageArgs
from ewoxcore.message.message_adapter import MessageAdapter

#C = TypeVar("C", redis.RedisCluster, redis.Redis)
C = TypeVar("C", RedisCluster, Redis)


class MessageStreamConsumer():
    def __init__(self, settings:ConsumerSettings, client:IRedisClient, consumer:IMessageConsumer) -> None:
        self._settings:ConsumerSettings = settings
        self._client:IRedisClient = client
        self._consumer:IMessageConsumer = consumer


    async def subscribe(self) -> None:
        """ Subscribe to the message stream. """
        connection:C = self._client.get_connection()
        if not connection:
            raise Exception("No Redis connection available.")

        await self._create_group(connection)

        while True:
            try:
                await self._process(connection)
            # except RetryError as re:
            #     print(f"Error: Sending to dead-letter queue.")
            #     connection.xadd(DEAD_STREAM, data)
            #     connection.xack(self._settings.stream_name, self._settings.group_name, mid)
            except Exception as e:
                print("[!] Unexpected error:", e)
                await asyncio.sleep(2)


    async def _create_group(self, connection:C) -> None:
        try:
            await connection.xgroup_create(name=self._settings.stream_name, groupname=self._settings.group_name, id="0", mkstream=True)
            print(f"Created consumer group '{self._settings.group_name}' on stream '{self._settings.stream_name}'")
#        except redis.ResponseError as e:
        except ResponseError as e:
            if "BUSYGROUP" in str(e):
                print(f"Consumer group '{self._settings.group_name}' already exists.")
            else:
                raise


    @retry(stop=stop_after_attempt(3), wait=wait_fixed(2))
    async def process_message(self, msg_id:str, msg_data:Dict[str, Any]) -> None:
        print(f"Processing message {msg_id}: {str(msg_data)}")

        args:MessageArgs = MessageAdapter.parse(msg_data)

        await self._consumer.on_consume(
            command=args.command,
            correlation_id=args.correlationId,
            json64=args.data,
            service_name=args.serviceName)


    @retry(
        wait=wait_exponential(min=1, max=10),
        stop=stop_after_attempt(5),
        retry_error_callback=lambda retry_state: None,
        retry=lambda retry_state: isinstance(
            retry_state.outcome.exception(),
            (ConnectionError, TimeoutError, OSError),
        )
    )    
    async def _process(self, connection:C) -> None:
        results = await connection.xreadgroup(
            groupname=self._settings.group_name,
            consumername=self._settings.consumer_name,
            streams={self._settings.stream_name: '>'},
            count=self._settings.count,
            block=5000
        )

        if results:
            for stream, messages in results:
                for msg_id, msg_data in messages:
                    try:
                        await self.process_message(msg_id, msg_data)
                        await self._ack_message(connection, msg_id)
                    except Exception as e:
                        print(f"Failed to process message {msg_id}: {e}")
                        
                        # Write to dead letter stream
                        await connection.xadd(
                            f"{self._settings.stream_name}:deadletter",
                            fields={
                                "original_id": msg_id,
                                "data": str(msg_data),
                                "error": str(e),
                            },
                            maxlen=10000,  # keep only latest 10,000 entries
                            approximate=True
                        )                        
                        # Ack the original message
                        await self._ack_message(connection, msg_id)
        else:
            # If no messages were processed, check for pending messages to claim
            await self._claim_pending(connection)


    async def _ack_message(self, connection:C, msg_id:str) -> None:
        """ Acknowledge a message in the stream. """
        await connection.xack(self._settings.stream_name, self._settings.group_name, msg_id)
        if (self._settings.deleteAfterAck):
            await connection.xdel(self._settings.stream_name, msg_id)


    async def _claim_pending(self, connection:C) -> None:
        """ Claim pending messages from failed consumers older than X seconds. """
        claim_after_ms: int = self._settings.claim_after_seconds * 1000
        start_id = "0-0"

        while True:
            next_id, claimed, *_ = await connection.xautoclaim(
                self._settings.stream_name,
                self._settings.group_name,
                self._settings.consumer_name,
                min_idle_time=claim_after_ms,
                start_id=start_id,
                count=100,
            )

            for mid, data in claimed:
                await self.process_message(mid, data)
                await connection.xack(self._settings.stream_name, self._settings.group_name, mid)

            if not claimed:
                break

            start_id = next_id


"""
import asyncio
from contextlib import suppress

class Consumer:
    def __init__(self, settings, max_concurrency: int = 20):
        self._settings = settings
        self._sem = asyncio.Semaphore(max_concurrency)

    async def _handle_one(self, connection, msg_id, msg_data):
        # Run one message under the concurrency gate
        async with self._sem:
            try:
                await self.process_message(msg_id, msg_data)
                self._ack_message(connection, msg_id)
            except Exception as e:
                # Log + dead-letter then ack original, same as before
                print(f"Failed to process message {msg_id}: {e}")
                connection.xadd(
                    f"{self._settings.stream_name}:deadletter",
                    fields={
                        "original_id": msg_id,
                        "data": str(msg_data),
                        "error": str(e),
                    },
                    maxlen=10000,
                    approximate=True,
                )
                self._ack_message(connection, msg_id)

    async def _process(self, connection) -> None:
        #Read in batches; process each message concurrently.
        while True:
            # If you're using redis.asyncio, this call should be awaited.
            # If it's a sync client, it will block the loop — consider switching to redis.asyncio (see note below).
            results = connection.xreadgroup(
                groupname=self._settings.group_name,
                consumername=self._settings.consumer_name,
                streams={self._settings.stream_name: '>'},
                count=10,
                block=5000,  # 5s long-poll
            )

            if results:
                tasks = []
                for stream, messages in results:
                    for msg_id, msg_data in messages:
                        tasks.append(asyncio.create_task(self._handle_one(connection, msg_id, msg_data)))

                # Let the whole batch run concurrently; exceptions are handled in _handle_one
                # so we don't need return_exceptions=True.
                # If you prefer to keep reading new batches while the current batch runs,
                # you can omit this await (but then track outstanding tasks to avoid unbounded growth).
                await asyncio.gather(*tasks)
            else:
                # No new messages — try to claim pending
                await self._claim_pending(connection)

"""


from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
from abc import ABC, abstractmethod
import redis
from ewoxcore.message.message_args import MessageArgs

T = TypeVar("T", redis.RedisCluster, redis.Redis)


class IMessageStreamPublisher(ABC):
    @abstractmethod
    async def publish(self, model:T) -> bool:
        """ Publish a message to the stream. """
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def publish_command(self, command:str, model:T) -> bool:
        """ Publish a command to the stream. """
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def publish_message(self, args:MessageArgs) -> bool:
        """ Publish a message to the stream. """
        raise NotImplementedError("Implement inhereted method")



from typing import Any, List, Dict, Text, Optional, Tuple, TypeVar, Union, Awaitable
import os
import time
import logging
import redis
from redis.backoff import ExponentialBackoff
from redis.retry import Retry
from ewoxcore.constants.server_env import ServerEnv
from ewoxdbredis.settings.connection_settings import ConnectionSettings

T = TypeVar("T", redis.RedisCluster, redis.Redis)


class DBConnection():
    def __init__(self, settings:ConnectionSettings=ConnectionSettings()) -> None:
        self._settings:ConnectionSettings = settings
        self._connection:T = None


    """ Enter & Exit to use together with the WITH statement. """
    def __enter__(self):
        self._connection = self._connect()
        return self._connection


    def __exit__(self, exc_type, exc_value, tb):
        self._connection.close()


    def _connect(self) -> Union[redis.RedisCluster, redis.Redis]:
        self._connection = self._connect_cluster() if (self._settings.is_cluster) else self._connect_single()
        return self._connection


    def _connect_single(self, num_retries:int=3) -> redis.Redis:
        client:redis.Redis = None

        try:
            client = redis.Redis(host=self._settings.host, port=self._settings.port, 
                                 retry=Retry(ExponentialBackoff(), retries=5))
        except Exception as error:
            logging.error(f"RedisClient.Process. numRetries: {str(num_retries)}, Error: {error}")
            if (num_retries > 0):
                time.sleep(15)
                num_retries -= 1
                return self._connect_single(num_retries)

        return client


    def _connect_cluster(self, num_retries:int=3) -> redis.RedisCluster:
        rc:redis.RedisCluster = None

        try:
            rc = redis.RedisCluster(host=self._settings.host, port=self._settings.port, ssl=True)
        except Exception as error:
            logging.error(f"RedisClient.Process. numRetries: {str(num_retries)}, Error: {error}")
            if (num_retries > 0):
                time.sleep(15)
                num_retries -= 1
                return self._connect_cluster(num_retries)

        return rc



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type, cast
import redis
import asyncio
from redis.exceptions import ConnectionError, TimeoutError
from tenacity import retry, wait_exponential, wait_fixed, stop_after_attempt, RetryError
from ewoxcore.utils.json_util import JsonUtil
from ewoxdbredis.clients.iredis_client import IRedisClient
from ewoxdbredis.settings.publisher_settings import PublisherSettings
from ewoxdbredis.stream.imessage_stream_publisher import IMessageStreamPublisher
from ewoxcore.message.message_args import MessageArgs

T = TypeVar("T")
C = TypeVar("C", redis.asyncio.RedisCluster, redis.asyncio.Redis)


class MessageStreamPublisher(IMessageStreamPublisher):
    def __init__(self, settings:PublisherSettings, client:IRedisClient) -> None:
        self._settings:PublisherSettings = settings
        self._client:IRedisClient = client


    async def publish(self, model:T) -> bool:
        """ Publish a message to the stream. """
        encoded_data:str = JsonUtil.serializeJson64(model)
        args:MessageArgs = MessageArgs(encoded_data)
        res:bool = await self.publish_message(args)

        return res


    async def publish_command(self, command:str, model:T) -> bool:
        """ Publish a command to the stream. """
        encoded_data:str = JsonUtil.serializeJson64(model)
        args:MessageArgs = MessageArgs(encoded_data, command=command)
        res:bool = await self.publish_message(args)

        return res


    @retry(
        wait=wait_exponential(min=1, max=10),
        stop=stop_after_attempt(5),
        retry_error_callback=lambda retry_state: None,
        retry=lambda retry_state: isinstance(
            retry_state.outcome.exception(),
            (ConnectionError, TimeoutError, OSError)
        )
    )
    async def publish_message(self, args:MessageArgs) -> bool:
        """ Publish a message to the stream. """
        connection:C = self._client.get_connection()
        if not connection:
            print("No Redis connection available.")
            return False

        try:
            data:Dict[str, str] = {
                "data": args.data,
                "correlationId": args.correlationId,
                "command": args.command,
                "serviceName": args.serviceName,
                "serverName": args.serverName,
                "sendAt": args.sendAt.isoformat()
            }

            mid = await connection.xadd(self._settings.stream_name, data)
            print(f"Published message with ID: {mid}")
        except (ConnectionError, TimeoutError) as e:
            print(f"Error publishing message: {e}")
            return False
        except Exception as e:
            print(f"Unexpected error: {e}")
            return False

        return True
